#!/usr/bin/env python3
#Copyright (C) 2020 Aloril
#License: Apache-2.0
#Adds RMobilityResult to pgn file, usage: parallel_tag_rmobility.py input.pgn output.pgn
#See https://wiki.chessdom.org/R-Mobility
#Will use as many processes as reported by multiprocessing.cpu_count()
#Positions in KvK, KBvK and KNvK *not* scored using r-mobility tablebase
import sys, chess.pgn, collections, multiprocessing

PROCESSES = multiprocessing.cpu_count()
INF = 1000

class Visitor(chess.pgn.BaseVisitor):
    def begin_game(self):
        self.reset()
        self.ply = -1

    def reset(self):
        self.bestPosVal = INF
        self.bestPosValPly = None

    def visit_board(self, board):
        self.ply += 1
        if self.ply==0 and board.turn!=chess.WHITE:
            self.ply += 1
        if board.halfmove_clock==0:
            self.reset()
        count = len(list(board.legal_moves))
        posVal = count
        if not board.is_check():
            posVal += 0.5
        if posVal < self.bestPosVal and (posVal < 1 or board.halfmove_clock < 100):
            self.bestPosVal = posVal
            self.bestPosValPly = self.ply

    def end_game(self):
        #1-0 and 0-1 results are handled elsewhere
        if self.bestPosValPly%2==0:
            neg_s = "-"
        else:
            neg_s = " "
        self.goal_str = "%sG%.1f" % (neg_s.strip(), self.bestPosVal)

    def result(self):
        return self.goal_str

def score(input_queue, queue):
    with open(filename, encoding=file_endocing) as fp:
        while True:
            count, offset = input_queue.get()
            if offset < 0:
                break
            fp.seek(offset)
            pgn = chess.pgn.read_game(fp)
            if not pgn:
                break
            visitor = Visitor()
            visitor.offset = offset
            visitor.offset2 = fp.tell()
            try:
                pgn.accept(visitor)
                result = pgn.headers.get("Result")
                if result=="1-0":
                    visitor.goal_str = "G0.0"
                elif result=="0-1":
                    visitor.goal_str = "-G0.0"
                queue.put((count, visitor))
            except:
                print("Skipping game with error:", pgn.headers)
        queue.put((None, None))

def queue_scoring(filename, send_queue):
    count = 0
    with open(filename, encoding=file_endocing) as fp:        
        while True:
            offset = fp.tell()
            headers = chess.pgn.read_headers(fp)
            if headers is None:
                break
            send_queue.put((count, offset))
            count += 1
    for _ in range(PROCESSES): send_queue.put((-1, -1))
        
if __name__=="__main__":
    filename, output_filename = sys.argv[1:3]
    if open(filename, "rb").read(3)==b'\xef\xbb\xbf':
        file_endocing = "utf-8"
    else:
        file_endocing = "latin-1"
    q = multiprocessing.Queue()
    offset_queue = multiprocessing.Queue()
    p = multiprocessing.Pool(PROCESSES, score, (offset_queue, q))
    p2 = multiprocessing.Pool(1, queue_scoring, (filename, offset_queue))
    exit_count = 0
    next_count = 0
    process_dict = {}
    with open(filename, encoding=file_endocing) as fp_in:
        with open(output_filename, "w", encoding=file_endocing) as fp_out:
            while exit_count < PROCESSES:
                count, visitor = q.get()
                if count==None:
                    exit_count += 1
                    continue
                process_dict[count] = visitor
                while next_count in process_dict:
                    visitor = process_dict.pop(next_count)
                    next_count += 1
                    tag = '[RMobilityResult "%s"]\n' % visitor.goal_str
                    read_ok = False
                    if visitor.offset < visitor.offset2 < 2**64: #For some python versions and some inputs can't just subtract offsets
                        fp_in.seek(visitor.offset)
                        pgn = fp_in.read(visitor.offset2 - visitor.offset)
                        read_ok = fp_in.tell()==visitor.offset2
                    if not read_ok: #Fallback to reading by lines
                        fp_in.seek(visitor.offset)
                        l = []
                        while fp_in.tell() != visitor.offset2:
                            l.append(fp_in.readline())
                        pgn = "".join(l)
                    pos = pgn.find("[Result")
                    if pos < 0:
                        pos = pgn.find("[Event")
                    if pos > 0:
                        pos2 = pgn[pos+1:].find("[")
                        if pos2 > 0:
                            pos += 1 + pos2
                    if pos < 0:
                        print("Warning, something probably wrong with following game:")
                        print(pgn)
                        pgn = tag + pgn
                    else:
                        pgn = pgn[:pos] + tag + pgn[pos:]
                    fp_out.write(pgn)
